# --------------------------------------------------------------------------------------------------
# Script to manually generate a new portal access-key for local (localhost) development purposes.
# --------------------------------------------------------------------------------------------------
# This is really only needed until smaht-portal has been fleshed out enough to be able to do this
# normally using the UI; as of August 2023 there no way to do this. So we generate a new access-key,
# and associated secret, and either insert it directly into your locally running portal database,
# or output JSON suitable for doing this via master-inserts/access_key.json; and either update
# your access-keys file (~/.smaht-keys.json) directly, or output JSON suitable for this file.
#
# The --user arguments is used to specify the user with which the new access-key will be associated.
# This may be either an explicit UUID or your choce, or an email address which must be present in
# the master-inserts/user.json file; this is (only) required if --update-database is specified.
#
# With no other arguments this script outputs JSON objects suitable for inserting into the
# database (via master-inserts/access_key.json), and for placing in your ~/.smaht-keys.json file.
#
# If the --update-database option is given, then the new access-key will automatically be written
# to your locally running instance of the portal, which (obviously) needs to be up/running.
#
# If the --update-keys option is given, then your ~/.smaht-keys.json file will be automatically
# updated with the new access-key (for the "smaht-local" property).
#
# The --update option may be used to specify both --update-database and --update-keys.
#
# This is intended to be run from the smaht-portal repo; there is a generate-local-access-key
# poetry script defined there to invoke this module. And just for convenience, this is also
# supported in the cgap-portal and fourfront repos; to use it from from cgap-portal or
# fourfront use the --app cgap or --app fourfront arguments, respectively.
#
# Example command:
#  generate-local-access-key
#
# Example output:
#   Creating a new local portal access-key ... Done.
#   New local portal access-key record suitable for: /Users/dmichaels/.smaht-keys.json:
#   {
#       "key": "DHU74PRA",
#       "secret": "fptkcxqdqqpbenin",
#       "server": "http://localhost:8000"
#   }
#   New local portal access-key insert record suitable for: tests/data/master-inserts/access_key.json:
#   {
#       "status": "current",
#       "user": "<your-user-uuid>",
#       "description": "Manually generated local access-key for testing.",
#       "access_key_id": "DHU74PRA",
#       "secret_access_key_hash": "SampleHash_pcHx3KtfzvHCXTpcINCm0LRvzORXBpviBy6SSJyMssmcKrsmcKrzs",
#       "uuid": "c04987b2-42e6-47fc-8247-9de7e417355b"
#   }
#
# Example command:
#   generate-local-access-key --user david_michaels@hms.harvard.edu --update
#
# Example output:
#   Creating a new local portal access-key ... Done.
#   Writing new local portal access-key to: /Users/dmichaels/.smaht-keys.json ... Done.
#   Writing new local portal access-key to locally running portal database ... Done.
# --------------------------------------------------------------------------------------------------

import argparse
from collections import namedtuple
import configparser
from datetime import datetime
import dateutil.tz
import io
import json
import os
from passlib.context import CryptContext
from passlib.registry import register_crypt_handler
import pytz
import requests
from typing import Optional, Tuple
import uuid
from dcicutils.common import AnyJsonData
from dcicutils.portal_utils import Portal
from snovault.authentication import (
    generate_password as generate_access_key_secret,
    generate_user as generate_access_key
)
from snovault.edw_hash import EDWHash
from snovault.loadxl import load_all
from ..project_app import app_project
from .captured_output import captured_output

_INSERTS_DIR = "src/encoded/tests/data/master-inserts"
_USER_INSERTS_FILE = f"{_INSERTS_DIR}/user.json"
_DEFAULT_INI_FILE = "development.ini"


def main() -> None:

    parser = argparse.ArgumentParser(description="Create local portal access-key for dev/testing purposes.")
    parser.add_argument("--user", required=False,
                        help=f"User email for which the access-key should be defined (in master-inserts/user.json); or a UUID.")
    parser.add_argument("--update", action="store_true", required=False, default=False,
                        help=f"Same as --update-database and --update-keys both.")
    parser.add_argument("--update-database", action="store_true", required=False, default=False,
                        help=f"Updates the database of your locally running portal with the new access-key.")
    parser.add_argument("--update-keys", action="store_true", required=False, default=False,
                        help=f"Updates your access-keys file (e.g. ~/.smaht-keys.json) with the new access-key (e.g. smaht-local).")
    parser.add_argument("--port", type=int, required=False, default=8000,
                        help="Port for localhost on which your local portal is running.")
    parser.add_argument("--ini", type=str, required=False, default=_DEFAULT_INI_FILE,
                        help=f"Name of the application .ini file; default is: {_DEFAULT_INI_FILE}")
    parser.add_argument("--app", choices=["smaht", "cgap", "fourfront"], required=False, default=_guess_default_app(),
                        help="App name for which the access-key should be generated; default is smaht.")
    parser.add_argument("--list", action="store_true", required=False, default=False, help="Just list access-keys.")
    parser.add_argument("--verbose", action="store_true", required=False, default=False, help="Verbose output.")
    parser.add_argument("--debug", action="store_true", required=False, default=False, help="Debugging output.")
    args = parser.parse_args()

    _ACCESS_KEYS_FILE = os.path.expanduser(f"~/.{args.app}-keys.json")
    _ACCESS_KEYS_FILE_ITEM = f"{args.app}-local"

    if args.update:
        args.update_database = True
        args.update_keys = True

    if args.update_database or args.list:
        # Just FYI an absolute minimal .ini file looks like this:
        # [app:app]
        # use = egg:encoded
        # sqlalchemy.url = postgresql://postgres@localhost:5441/postgres?host=/tmp/snovault/pgdata
        # multiauth.groupfinder = encoded.authorization.smaht_groupfinder
        # multiauth.policies = auth0 session remoteuser accesskey
        # multiauth.policy.session.namespace = mailto
        # multiauth.policy.session.use = encoded.authentication.NamespacedAuthenticationPolicy
        # multiauth.policy.session.base = pyramid.authentication.SessionAuthenticationPolicy
        # multiauth.policy.remoteuser.namespace = remoteuser
        # multiauth.policy.remoteuser.use = encoded.authentication.NamespacedAuthenticationPolicy
        # multiauth.policy.remoteuser.base = pyramid.authentication.RemoteUserAuthenticationPolicy
        # multiauth.policy.accesskey.namespace = accesskey
        # multiauth.policy.accesskey.use = encoded.authentication.NamespacedAuthenticationPolicy
        # multiauth.policy.accesskey.base = encoded.authentication.BasicAuthAuthenticationPolicy
        # multiauth.policy.accesskey.check = encoded.authentication.basic_auth_check
        # multiauth.policy.auth0.use = encoded.authentication.NamespacedAuthenticationPolicy
        # multiauth.policy.auth0.namespace = auth0
        # multiauth.policy.auth0.base = encoded.authentication.Auth0AuthenticationPolicy
        with captured_output(not args.debug):
            portal = Portal(args.ini)
    else:
        portal = None

    if args.list:
        _print_all_access_keys(portal, args.verbose)
        return

    print(f"Creating a new local portal access-key for {args.app} ... ", end="")
    access_key_user_uuid = _generate_user_uuid(args.user, portal)
    access_key_id, access_key_secret, access_key_secret_hash = _generate_access_key(args.ini)
    access_key_inserts_file_item = _generate_access_key_inserts_item(access_key_id, access_key_secret_hash, access_key_user_uuid)
    access_keys_file_item = _generate_access_keys_file_item(access_key_id, access_key_secret, args.port)
    print("Done.")

    if args.update_keys:
        print(f"Writing new local portal access-key to: {_ACCESS_KEYS_FILE} ... ", end="")
        access_keys_file_json = {}
        try:
            with io.open(_ACCESS_KEYS_FILE, "r") as access_keys_file_f:
                access_keys_file_json = json.load(access_keys_file_f)
        except Exception:
            pass
        access_keys_file_json[_ACCESS_KEYS_FILE_ITEM] = access_keys_file_item
        with io.open(_ACCESS_KEYS_FILE, "w") as access_keys_file_f:
            json.dump(access_keys_file_json, fp=access_keys_file_f, indent=4)
        print("Done.")
    if not args.update_keys or args.verbose:
        print(f"New local portal access-key record suitable for: {_ACCESS_KEYS_FILE} ...")
        print(json.dumps(access_keys_file_item, indent=4))

    if args.update_database:
        if not _is_local_portal_running(args.port):
            _exit_without_action(f"Portal must be running locally ({_get_local_portal_url(args.port)}) to do an insert.")
        print(f"Writing new local portal access-key to locally running portal database ... ", end="")
        with captured_output(not args.debug):
            _load_data(portal, access_key_inserts_file_item, data_type="access_key")
        print("Done.")
    if not args.update_database or args.verbose:
        print(f"New local portal access-key insert record suitable for: {_INSERTS_DIR}/access_key.json ...")
        print(json.dumps(access_key_inserts_file_item, indent=4))


def _generate_user_uuid(user: Optional[str], portal: Optional[Portal] = None) -> str:
    if not user:
        if portal:
            _exit_without_action(f"The --user option must specify a UUID or email in {_USER_INSERTS_FILE}")
        return "<your-user-uuid>"
    user_uuid = None
    if _is_uuid(user):
        user_uuid = user
    else:
        with io.open(_USER_INSERTS_FILE, "r") as user_inserts_f:
            user_uuid_from_inserts = [item for item in json.load(user_inserts_f) if item.get("email") == user]
            if not user_uuid_from_inserts:
                _exit_without_action(f"The given user ({user}) was not found as an email"
                                     f" in: {_USER_INSERTS_FILE}; and it is not a UUID.")
            user_uuid = user_uuid_from_inserts[0]["uuid"]
    if user_uuid and portal:
        user = portal.get(f"/{user_uuid}", raise_exception=False)
        if not user or user.status_code != 200:
            _exit_without_action(f"The given user ({user_uuid}) was not found in the locally running portal database.")
    return user_uuid


def _generate_access_key_inserts_item(access_key_id: str, access_key_secret_hash: str, user_uuid: str) -> dict:
    return {
        "status": "current",
        "user": user_uuid,
        "description": f"Manually generated local access-key for testing.",
        "access_key_id": access_key_id,
        "secret_access_key_hash": access_key_secret_hash,
        "uuid": str(uuid.uuid4())
    }


def _generate_access_keys_file_item(access_key_id: str, access_key_secret: str, port: int) -> dict:
    return {
        "key": access_key_id,
        "secret": access_key_secret,
        "server": _get_local_portal_url(port)
    }


def _generate_access_key(ini_file: str = _DEFAULT_INI_FILE) -> Tuple[str, str, str]:
    access_key_secret = generate_access_key_secret()
    return generate_access_key(), access_key_secret, _hash_secret_like_snovault(access_key_secret, ini_file)


def _hash_secret_like_snovault(secret: str, ini_file: str = _DEFAULT_INI_FILE) -> str:

    # We do NOT store the secret in plaintext in the database, but rather a hash of it; this function
    # hashes the (given) secret in the same way that the portal (snovault) does and returns this result.
    # See access_key_add in snovault/types/access_key.py and includeme in snovault/authentication.py.
    # Using that code directly from snovault is a little tricker then we want to deal with for this;
    # and/but we do make an effort to read any passlib properties which might exist in the .ini file,
    # just like snovault does; perhaps overkill; default is development.ini; change with --ini. 
    def get_passlib_properties_from_ini_file(ini_file : str = _DEFAULT_INI_FILE,
                                             section_name = "app:app",
                                             property_name_prefix = "passlib.") -> str:
        """
        Returns from the specified section of the specified .ini file the values of properties with
        the specified property name prefix, in the form of a dictionary, where the property names have
        that specified property name prefix removed, and the property value is the associated value.
        """
        properties = {}
        try:
            config = configparser.ConfigParser()
            read_files = config.read(ini_file)
            if not read_files or read_files[0] != ini_file:
                _exit_without_action(f"The given ini file ({ini_file}) cannot be read.")
            for property_name in [p for p in config.options(section_name) if p.startswith(property_name_prefix)]:
                property_value = config.get(section_name, property_name)
                properties[property_name[len(property_name_prefix):]] = property_value
        except Exception:
            pass
        return properties

    passlib_properties = get_passlib_properties_from_ini_file(ini_file)
    if not passlib_properties:
        passlib_properties = {"schemes": "edw_hash, unix_disabled"}
    register_crypt_handler(EDWHash)
    return CryptContext(**passlib_properties).hash(secret)


def _guess_default_app() -> str:
    # This should return one of: smaht-portal, cgap-portal, or fourfront
    # that is, if running within repos: smaht-portal, cgap-portal, or fourfront
    repository = app_project().REPO_NAME
    return repository if repository != "cgap-portal" else "cgap"


def _is_local_portal_running(port: int) -> None:
    try:
        return requests.get(f"{_get_local_portal_url(port)}/health").status_code == 200
    except Exception:
        return False


def _get_local_portal_url(port: int) -> None:
    return f"http://localhost:{port}"


def _load_data(portal: Portal, data: AnyJsonData, data_type: str) -> bool:
    if isinstance(data, dict):
        data = [data]
    elif not isinstance(data, list):
        return False
    if not data_type:
        return False
    data = {data_type: data}
    load_all(portal.vapp, inserts=data, docsdir=None, overwrite=True, itype=[data_type], from_json=True)
    return True


def _print_all_access_keys(portal: Portal, verbose: bool = False) -> None:
    print("All access-keys defined for locally running portal:")
    for item in _get_all_access_keys(portal):
        print(f"{item.id}", end="")
        if item.created:
            print(f" | Created: {item.created}", end="")
        if item.expires:
            print(f" | Expires: {item.expires}", end="")
        if verbose:
            print(f" | {item.uuid}", end="")
        print()


def _get_all_access_keys(portal: Portal) -> list:
    response = []
    try:
        AccessKey = namedtuple("AccessKey", ["id", "uuid", "created", "expires"])
        for access_key in portal.get(f"/access-keys").json()["@graph"]:
            response.append(AccessKey(
                id=access_key.get("access_key_id"),
                uuid=access_key.get("uuid"),
                created=_format_iso_datetime_string_to_local_datetime_string(access_key.get("date_created")),
                expires=_format_iso_datetime_string_to_local_datetime_string(access_key.get("expiration_date"))))
    except Exception:
        pass
    return sorted(response, key=lambda item: item.created, reverse=True)


# TODO: Use misc_utils.is_uuid when out of review ...
# https://github.com/4dn-dcic/utils/blob/kmp_sheet_utils_schema_hinting/dcicutils/misc_utils.py#L1371
def _is_uuid(s: str) -> bool:
    try:
        return str(uuid.UUID(s)) == s
    except Exception:
        return False


def _exit_without_action(message: str) -> None:
    print(f"\nERROR: {message}")
    exit(1)


# TODO: Move these datetime utility function to dcicutils ...

def _parse_iso_datetime_string(value: str, format: str = "%Y-%m-%dT%H:%M:%S.%f") -> datetime:
    try:
        if value.endswith("+00:00"):
            value = value[:-6]
        return pytz.utc.localize(datetime.strptime(value, format))
    except Exception:
        return ""


def _convert_iso_to_local_datetime(value: datetime, format: str = "%Y-%m-%dT%H:%M:%S.%f") -> datetime:
    try:
        return value.astimezone(dateutil.tz.tzlocal())
    except Exception:
        return ""


def _format_datetime(value: datetime, format: str = "%Y-%m-%d %H:%M:%S %Z") -> str:
    try:
        return value.strftime(format)
    except Exception:
        return ""


def _format_iso_datetime_string_to_local_datetime_string(value: str) -> str:
    return _format_datetime(_convert_iso_to_local_datetime(_parse_iso_datetime_string(value)))


if __name__ == "__main__":
    main()



from uuid import UUID

from .schema_utils import validate_request, validate, IgnoreUnchanged
from .validation import ValidationFailure


# No-validation validators


def no_validate_item_content_post(context, request):
    data = request.json
    request.validated.update(data)


def no_validate_item_content_put(context, request):
    data = request.json
    if 'uuid' in data:
        if UUID(data['uuid']) != context.uuid:
            msg = 'uuid may not be changed'
            raise ValidationFailure('body', ['uuid'], msg)
    request.validated.update(data)


def no_validate_item_content_patch(context, request):
    data = context.properties.copy()
    data.update(request.json)
    schema = context.type_info.schema
    # this will raise a validation error if delete_fields param provided
    data = delete_fields(request, data, schema)
    if 'uuid' in data:
        if UUID(data['uuid']) != context.uuid:
            msg = 'uuid may not be changed'
            raise ValidationFailure('body', ['uuid'], msg)
    request.validated.update(data)


def delete_fields(request, data, schema):
    """
    Delete fields from data in the delete_fields param of the request.
    Validate to catch any permission errors and return the validated data

    Args:
        request: current Request object
        data: dict item contents
        schema: dict item schema

    Returns:
        dict validated contents

    Raises:
        ValidationFailure if validate=false in request params
    """
    if not request.params.get('delete_fields'):
        return data

    # do not allow validate=false with delete_fields, since it skips schema
    # validation, which does things like adding defaults
    if request.params.get('validate') == 'false':
        err_msg = 'Cannot delete fields on request with with validate=false'
        raise ValidationFailure('body', 'delete_fields', err_msg)

    # make a copy of data after removing fields and validate
    to_validate = data.copy()
    for dfield in request.params['delete_fields'].split(','):
        dfield = dfield.strip()
        if dfield in data:
            del to_validate[dfield]
    # permission validation, comparing data with deleted values to the previous
    # data. If validated, return that validated value
    # Note 1: a deleted field with a default will be replaced with that value;
    #         if same value was already in data, allow regardless of permission
    # Note 2: setting validate_current = False here creates a vulnerability in
    # that non-admin users can use delete_fields to revert protected fields to
    # their defaults - see PR 95 - Will June 29 2021
    # TODO: validate edit permission on query fields, do not validate current
    validated, errors = validate(schema, to_validate, current=data, validate_current=True)
    if errors:
        for error in errors:
            if isinstance(error, IgnoreUnchanged):
                if error.validator != 'permission':
                    continue
            error_name = 'Schema: ' + '.'.join([str(p) for p in error.path])
            request.errors.add('body', error_name, error.message)
    if request.errors:
        raise ValidationFailure('body', 'delete_fields', 'Error deleting fields')
    # validate() may add fields, such as uuid and schema_version
    for orig_field in [k for k in validated]:
        if orig_field not in to_validate:
            del validated[orig_field]
    # finally, return the validated contents with deleted fields
    return validated


# Schema checking validators
def validate_item_content_post(context, request):
    data = request.json
    validate_request(context.type_info.schema, request, data)


def validate_item_content_put(context, request):
    data = request.json
    schema = context.type_info.schema
    if 'uuid' in data and UUID(data['uuid']) != context.uuid:
        msg = 'uuid may not be changed'
        raise ValidationFailure('body', ['uuid'], msg)
    current = context.upgrade_properties().copy()
    current['uuid'] = str(context.uuid)
    validate_request(schema, request, data, current)


def validate_item_content_patch(context, request):
    data = context.upgrade_properties().copy()
    if 'schema_version' in data:
        del data['schema_version']
    data.update(request.json)
    schema = context.type_info.schema
    data = delete_fields(request, data, schema)
    if 'uuid' in data and UUID(data['uuid']) != context.uuid:
        msg = 'uuid may not be changed'
        raise ValidationFailure('body', ['uuid'], msg)
    current = context.upgrade_properties().copy()
    current['uuid'] = str(context.uuid)
    # this will add defaults values back to deleted fields with schema defaults
    validate_request(schema, request, data, current)


def validate_item_content_in_place(context, request):
    """
    Used with indexer. Validate the request against the schema using only the
    data in the request, which will be upgraded context properties, to save a
    call to the DB with `context.upgrade_properties`

    Args:
        context: current Item context
        request: current Request

    Returns:
        None

    Raises:
        ValidationFailure: on uuid mismatch or schema validation failure
    """
    data = request.json
    schema = context.type_info.schema
    if 'uuid' in data and UUID(data['uuid']) != context.uuid:
        msg = 'uuid may not be changed'
        raise ValidationFailure('body', ['uuid'], msg)
    # set current to data, since this validator is not meant to be used for
    # changing properties. Used with call to `schema_utils.validate`
    current = data.copy()
    current['uuid'] = str(context.uuid)
    validate_request(schema, request, data, current)



from __future__ import annotations

import argparse
import itertools
import json
import logging
import re
import structlog
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, Iterable, Iterator, List, Optional, Set, Union

from dcicutils import creds_utils
from dcicutils.ff_utils import get_metadata, search_metadata
from dcicutils.misc_utils import to_snake_case

from snovault.commands.utils import get_auth_key


logger = structlog.getLogger(__name__)
EPILOG = __doc__

DEFAULT_IGNORE_FIELDS = [
    "submitted_by",
    "date_created",
    "last_modified",
    "schema_version",
]
INSERT_DIRECTORIES = [
    "inserts",
    "master-inserts",
    "perf-testing",
    "workbook-inserts",
    "temp-local-inserts",
    "deploy-inserts",
]
INSERTS_LOCATION = Path("src/encoded/tests/data/")
KEYS_TO_IGNORE_FOR_LINKS = set(  # Either UUIDs not needed or can look like UUIDs
    ["uuid", "title", "blob_id", "md5sum", "content_md5sum", "submitted_md5sum"]
)
UUID_IDENTIFIER = re.compile(
    r"^[a-f0-9]{8}-?[a-f0-9]{4}-?4[a-f0-9]{3}-?[a-f0-9]{4}-?[a-f0-9]{12}\Z",
    re.I,
)


def get_ignore_fields(ignore_fields: List[str]) -> Set[str]:
    """Get all fields to ignore when pulling inserts."""
    return set(DEFAULT_IGNORE_FIELDS + ignore_fields)


def update_inserts_from_server(
    inserts: Path,
    auth_key: Dict[str, str],
    ignore_fields: Iterable[str],
    item_types: Optional[List[str]] = None,
    from_search: Optional[str] = None,
    merge_existing: bool = True,
) -> None:
    """Update inserts for given server."""
    existing_inserts_to_update = get_existing_inserts_to_update(inserts, item_types)
    if existing_inserts_to_update:
        logger.info(
            f"Found {len(existing_inserts_to_update)} existing inserts to update"
        )
    else:
        logger.info("No existing inserts to update")
    search_uuids = get_uuids_from_search(from_search, auth_key)
    if search_uuids:
        logger.info(f"Found {len(search_uuids)} items from search")
    base_uuids = get_base_uuids(existing_inserts_to_update, search_uuids)
    logger.info("Collecting inserts from portal. This may take a while...")
    inserts_from_portal = get_inserts_from_portal(
        base_uuids, auth_key, ignore_fields
    )
    logger.info(f"Found inserts for {len(inserts_from_portal)} item types from portal")
    inserts_to_write = get_inserts_to_write(
        inserts_from_portal, existing_inserts_to_update, merge_existing=merge_existing
    )
    logger.info(f"Writing inserts for {len(inserts_to_write)} item types to {inserts}")
    write_inserts(inserts_to_write, inserts)


@dataclass(frozen=True)
class Insert:
    uuid: str
    properties: Dict[str, Any]

    def __hash__(self):
        return hash(self.uuid)

    def __eq__(self, other: Any):
        if not isinstance(other, Insert):
            return False
        return self.uuid == other.uuid

    def __repr__(self):
        return f"Insert({self.uuid})"

    def update(
        self,
        properties: Optional[Dict[str, Any]] = None,
    ) -> Insert:
        """Update insert attributes."""
        if not properties or properties == self.properties:
            return self
        return Insert(
            uuid=self.uuid,
            properties=properties,
        )


@dataclass(frozen=True)
class ItemTypeInserts:
    item_type: str
    uuids: Set[str]
    inserts: Iterator[Insert]

    def __repr__(self):
        return f"{self.__class__.__name__}({self.item_type})"

    def add_insert(self, insert: Insert) -> ItemTypeInserts:
        """Update inserts for a given item type."""
        if insert.uuid in self.uuids:
            return self
        return ItemTypeInserts(
            item_type=self.item_type,
            uuids=self.uuids | {insert.uuid},
            inserts=itertools.chain(self.inserts, [insert]),
        )


def get_existing_inserts_to_update(
    inserts_path: Path,
    target_types: Optional[List[str]] = None,
) -> List[ItemTypeInserts]:
    """Get all existing inserts to update from given directory.

    Note: Pulling existing inserts into memory
    """
    existing_inserts = get_existing_inserts(inserts_path)
    return get_inserts_to_update(existing_inserts, target_types)


def get_existing_inserts(inserts_path: Path) -> List[ItemTypeInserts]:
    """Get all existing inserts from given directory."""
    inserts_files = inserts_path.glob("*.json")
    return [get_item_type_inserts(insert_file) for insert_file in inserts_files]


def get_item_type_inserts(insert_file: Path) -> ItemTypeInserts:
    """Get all inserts for a given item type.

    Ensure inserts are sorted by UUID for later comparison with portal
    inserts.
    """
    item_type = insert_file.stem
    uuids = set()
    inserts = []
    file_inserts = get_inserts_from_file(insert_file)
    for insert in sort_inserts_by_uuid(file_inserts):
        uuids |= {insert.uuid}
        inserts = itertools.chain(inserts, [insert])
    return ItemTypeInserts(
        item_type=item_type,
        uuids=uuids,
        inserts=inserts,
    )


def create_insert(insert: Dict[str, Any]) -> Insert:
    """Create Insert from JSON."""
    return Insert(
        uuid=get_uuid(insert),
        properties=insert,
    )


def get_inserts_from_file(insert_file: Path) -> Iterator[Insert]:
    """Load inserts from file."""
    with insert_file.open("r") as f:
        inserts = json.load(f)
    return (create_insert(insert) for insert in inserts)


def sort_inserts_by_uuid(inserts: Iterable[Insert]) -> List[Insert]:
    """Sort all inserts by UUID."""
    return sorted(inserts, key=lambda insert: insert.uuid)


def get_uuid(item: Dict[str, Any]) -> str:
    return item.get("uuid", "")


def get_inserts_to_update(
    existing_inserts: List[ItemTypeInserts],
    target_types: Optional[List[str]] = None,
) -> Dict[str, Iterator[Insert]]:
    """Get all inserts to update from given directory."""
    if target_types:
        item_types_to_keep = set([to_snake_case(target) for target in target_types])
        return [
            item_type_inserts
            for item_type_inserts in existing_inserts
            if item_type_inserts.item_type in item_types_to_keep
        ]
    return existing_inserts


def get_uuids_from_search(
    search_query: Union[str, None], auth_key: Dict[str, str]
) -> List[str]:
    """Get all uuids from a given search query."""
    if not search_query:
        return []
    query = format_search_query(search_query)
    return [get_uuid(item) for item in search_metadata(query, key=auth_key)]


def format_search_query(search_query: str) -> str:
    """Format provided search query.

    Include UUID add-on to limit response size and use heuristics to
    format query as needed (e.g. so no need to add boilerplate start of
    "search/").
    """
    if "field=uuid" not in search_query:
        search_query += "&field=uuid"
    if "/" in search_query:  # Likely formatted appropriately
        return search_query
    if not search_query.startswith("?"):
        return f"search/?{search_query}"
    return f"search/{search_query}"


def get_base_uuids(
    existing_inserts: List[ItemTypeInserts],
    search_uuids: List[str],
) -> Set[str]:
    """Get all base uuids to get from given existing inserts and search uuids."""
    existing_uuids = set().union(
        *[item_type_inserts.uuids for item_type_inserts in existing_inserts]
    )
    return existing_uuids | set(search_uuids)


def get_inserts_from_portal(
    uuids: Iterable[str], auth_key: Dict[str, str], ignore_fields: Set[str]
) -> List[ItemTypeInserts]:
    """Get all inserts to write.

    Start from base UUIDs and branch to all linked items.
    """
    item_type_inserts = []
    seen = set()
    uuids_to_get = set(uuids)
    while uuids_to_get:
        current_uuids_to_get = uuids_to_get - seen
        uuids_to_get = set()
        for uuid in current_uuids_to_get:
            item_type = get_item_type(uuid, auth_key)
            insert = get_insert(uuid, auth_key, ignore_fields)
            item_type_inserts = update_item_type_inserts(
                item_type_inserts, item_type, insert
            )
            seen |= {uuid}
            uuids_to_get |= get_links(insert) - seen
    return item_type_inserts


def get_item(uuid: str, auth_key: Dict[str, str], frame: str = "raw") -> Dict[str, Any]:
    """Get item for a given UUID."""
    add_on = f"frame={frame}" if frame else ""
    return get_metadata(uuid, key=auth_key, add_on=add_on)


def get_insert(uuid: str, auth_key: Dict[str, str], ignore_fields: Set[str]) -> Insert:
    """Get insert for a given item."""
    item = get_item(uuid, auth_key)
    properties = get_insert_properties(item, ignore_fields)
    return create_insert(properties)


def get_item_type(uuid: str, auth_key: Dict[str, str]) -> str:
    """Get item type for a given item.

    Ensure snake_cased to match insert file names.
    """
    item = get_item(uuid, auth_key, frame="object")
    return to_snake_case(item["@type"][0])


def get_insert_properties(
    item: Dict[str, Any], ignore_fields: Set[str]
) -> Dict[str, Any]:
    """Get all properties for a given item."""
    return {key: value for key, value in item.items() if key not in ignore_fields}


def update_item_type_inserts(
    item_type_inserts: List[ItemTypeInserts],
    item_type: str,
    insert: Insert,
) -> List[ItemTypeInserts]:
    """Update item type inserts with new insert.

    If item type already exists, add insert to existing item type inserts.
    Otherwise, create new item type inserts.
    """
    existing_item_types = {
        item_type_inserts.item_type for item_type_inserts in item_type_inserts
    }
    if item_type in existing_item_types:
        return [
            item_type_insert.add_insert(insert)
            if item_type_insert.item_type == item_type
            else item_type_insert
            for item_type_insert in item_type_inserts
        ]
    new_item_type_inserts = ItemTypeInserts(
        item_type=item_type,
        uuids={insert.uuid},
        inserts=(insert,),
    )
    return item_type_inserts + [new_item_type_inserts]


def get_links(item: Any) -> Set[str]:
    """Get all links for a given item."""
    if isinstance(item, Insert):
        return get_links_from_insert(item)
    if isinstance(item, dict):
        return get_links_from_dict(item)
    if isinstance(item, list):
        return get_links_from_list(item)
    if isinstance(item, str):
        if is_uuid(item):
            return {item}
    return set()


def get_links_from_insert(item: Insert) -> Set[str]:
    """Get all links for a given insert."""
    return get_links_from_dict(item.properties)


def get_links_from_dict(item: Dict[str, Any]) -> Set[str]:
    """Get all links for a given dictionary."""
    return {
        link
        for key, value in item.items()
        if key not in KEYS_TO_IGNORE_FOR_LINKS
        for link in get_links(value)
    }


def get_links_from_list(item: List[Any]) -> Set[str]:
    """Get all links for a given list."""
    return {link for value in item for link in get_links(value)}


def is_uuid(value: str) -> bool:
    """Check if a given value is a UUID."""
    return bool(UUID_IDENTIFIER.match(value))


def get_inserts_to_write(
    inserts_from_portal: List[ItemTypeInserts],
    existing_inserts_to_update: List[ItemTypeInserts],
    merge_existing: bool = True,
) -> List[ItemTypeInserts]:
    """Get all inserts to write.

    Update portal inserts with existing inserts information, if present,
    and remove conflicts with master-inserts.
    """
    if merge_existing:
        inserts_updated_with_existing = get_inserts_with_existing_data(
            inserts_from_portal, existing_inserts_to_update
        )
        return get_inserts_without_conflicts(inserts_updated_with_existing)
    return get_inserts_without_conflicts(inserts_from_portal)


def get_inserts_with_existing_data(
    inserts_from_portal: List[ItemTypeInserts],
    existing_inserts_to_update: List[ItemTypeInserts],
) -> List[ItemTypeInserts]:
    """Update portal inserts with existing inserts data, if present."""
    if not existing_inserts_to_update:
        return inserts_from_portal
    existing_item_types_to_inserts = {
        item_type_inserts.item_type: item_type_inserts
        for item_type_inserts in existing_inserts_to_update
    }
    return [
        get_updated_item_type_inserts(
            portal_item_type_inserts,
            existing_item_types_to_inserts[portal_item_type_inserts.item_type],
        )
        if are_item_type_inserts_present_and_overlapping(
            portal_item_type_inserts, existing_item_types_to_inserts
        )
        else portal_item_type_inserts
        for portal_item_type_inserts in inserts_from_portal
    ]


def are_item_type_inserts_present_and_overlapping(
    item_type_inserts: ItemTypeInserts,
    comparison_inserts: Dict[str, ItemTypeInserts],
) -> bool:
    """Check if item type inserts are present in and overlap inserts."""
    return (
        item_type_inserts.item_type in comparison_inserts
        and do_inserts_overlap(
            item_type_inserts, comparison_inserts[item_type_inserts.item_type]
        )
    )


def do_inserts_overlap(
    item_type_inserts_1: ItemTypeInserts,
    item_type_inserts_2: ItemTypeInserts,
) -> bool:
    """Check if portal inserts overlap with existing inserts."""
    return bool(item_type_inserts_1.uuids & item_type_inserts_2.uuids)


def get_updated_item_type_inserts(
    inserts_from_portal: ItemTypeInserts,
    existing_inserts: ItemTypeInserts,
) -> ItemTypeInserts:
    """Update portal item type inserts with existing inserts data.

    Note: Existing inserts are assumed to be sorted by UUID (performed
    when loaded from file).
    """
    sorted_portal_inserts = sort_inserts_by_uuid(inserts_from_portal.inserts)
    updated_inserts = (
        get_updated_insert(insert, existing_inserts.inserts)
        if insert.uuid in existing_inserts.uuids
        else insert
        for insert in sorted_portal_inserts
    )
    return ItemTypeInserts(
        item_type=inserts_from_portal.item_type,
        uuids=inserts_from_portal.uuids,
        inserts=updated_inserts,
    )


def get_updated_insert(insert: Insert, existing_inserts: Iterator[Insert]) -> Insert:
    """Update portal insert with data from existing insert.

    Run through existing inserts (presumed to be sorted by UUID) until
    match found for given insert (also presumed to be sorted by UUID).
    """
    for existing_insert in existing_inserts:
        if existing_insert.uuid == insert.uuid:
            return insert.update(existing_insert.properties)
    return insert


def get_inserts_without_conflicts(
    item_type_inserts: List[ItemTypeInserts],
) -> List[ItemTypeInserts]:
    """Remove all conflicts with master-inserts."""
    master_inserts = get_existing_inserts(INSERTS_LOCATION.joinpath("master-inserts"))
    master_inserts_item_types = {
        item_type_inserts.item_type: item_type_inserts
        for item_type_inserts in master_inserts
    }
    return [
        get_inserts_without_conflicts_for_item_type(
            item_type_inserts, master_inserts_item_types[item_type_inserts.item_type]
        )
        if are_item_type_inserts_present_and_overlapping(
            item_type_inserts, master_inserts_item_types
        )
        else item_type_inserts
        for item_type_inserts in item_type_inserts
    ]


def get_inserts_without_conflicts_for_item_type(
    portal_item_type_inserts: ItemTypeInserts,
    master_item_type_inserts: ItemTypeInserts,
) -> ItemTypeInserts:
    """Remove conflicts with master-inserts for a given item type.

    Assumes both portal and master inserts are sorted by UUID.
    """
    uuids = set()
    inserts = []
    for insert in sort_inserts_by_uuid(portal_item_type_inserts.inserts):
        if is_insert_in_master_inserts(insert, master_item_type_inserts):
            continue
        uuids |= {insert.uuid}
        inserts = itertools.chain(inserts, [insert])
    return ItemTypeInserts(
        item_type=portal_item_type_inserts.item_type,
        uuids=uuids,
        inserts=inserts,
    )


def is_insert_in_master_inserts(
    insert: Insert,
    master_item_type_inserts: ItemTypeInserts,
) -> bool:
    """Check if a given insert is in master-inserts.

    Assumes master inserts are sorted by UUID and portal inserts arrive
    in order as well.
    """
    if insert.uuid in master_item_type_inserts.uuids:
        for master_insert in master_item_type_inserts.inserts:
            if master_insert.uuid == insert.uuid:
                if master_insert.properties == insert.properties:
                    logging.info(
                        f"Skipping {insert.uuid} for"
                        f" {master_item_type_inserts.item_type} as"
                        f" already in master-inserts"
                    )
                else:
                    logging.warning(
                        f" Skipping {insert.uuid} for"
                        f" {master_item_type_inserts.item_type} as"
                        f" conflicts with master-inserts"
                    )
                return True
    return False


def write_inserts(item_type_inserts: List[ItemTypeInserts], inserts_path: Path) -> None:
    """Write all inserts to given directory."""
    for item_type_insert in item_type_inserts:
        write_inserts_for_type(
            item_type_insert.item_type, item_type_insert.inserts, inserts_path
        )


def write_inserts_for_type(
    item_type: str,
    inserts_for_type: Iterable[Insert],
    inserts_path: Path,
) -> None:
    """Write all inserts for a given item type to given directory."""
    insert_file = inserts_path.joinpath(f"{item_type}.json")
    inserts = sort_inserts_by_uuid(inserts_for_type)
    to_write = [insert.properties for insert in inserts]
    with insert_file.open("w") as file_handle:
        json.dump(to_write, file_handle, indent=4)
    logger.info(f"Wrote {len(inserts)} inserts for {item_type} to {insert_file}")


def main():
    """Update the inserts from a given portal.

    Use `--item-type` to update existing inserts for specific item types.
    Use `--from-search` to update inserts from a search result.
    """
    logging.basicConfig()
    logging.getLogger("encoded").setLevel(logging.DEBUG)

    parser = argparse.ArgumentParser(
        description="Update Inserts",
        epilog=EPILOG,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "--env", default="data", help="Environment to update from. Defaults to data"
    )
    parser.add_argument(
        "--portal",
        help="Portal to update inserts from. Defaults to SMaHT",
        choices=creds_utils._KEY_MANAGERS.keys(),
        default="smaht",
    )
    parser.add_argument(
        "--dest",
        default="temp-local-inserts",
        help="Destination inserts directory. Defaults to temp-local-inserts.",
    )
    parser.add_argument(
        "--item-type",
        nargs="+",
        help=(
            "Existing item type (e.g. file_fastq) to update inserts for."
            " Defaults to all types found in destination directory.",
        ),
    )
    parser.add_argument(
        "--ignore-field",
        nargs="+",
        default=DEFAULT_IGNORE_FIELDS,
        help="Properties to ignore when pulling inserts",
    )
    parser.add_argument(
        "--from-search",
        help="Query to find new items to add to inserts",
        type=str,
    )
    parser.add_argument(
        "--refresh",
        help=(
            "Replace existing inserts with new inserts from portal, i.e. don't keep"
            " properties from existing inserts not present in portal."
            " Defaults to False.",
        ),
        action="store_true",
    )
    args = parser.parse_args()

    ignore_fields = get_ignore_fields(args.ignore_field)
    auth_key = get_auth_key(args.portal, args.env)
    if args.dest not in INSERT_DIRECTORIES:
        proceed = input(
            f"Destination {args.dest} not in common choices: {INSERT_DIRECTORIES}."
            f"\nContinue? (y/n): "
        )
        if not proceed.lower().startswith("y"):
            logger.info("Exiting.")
            return
    inserts_path = INSERTS_LOCATION.joinpath(args.dest)
    if not inserts_path.exists():
        inserts_path.mkdir()
    update_inserts_from_server(
        inserts_path,
        auth_key,
        ignore_fields,
        item_types=args.item_type,
        from_search=args.from_search,
        merge_existing=not args.refresh,
    )


if __name__ == "__main__":
    main()



# Renderers related functions which may be overriden by an implementing app,
# e.g. Foursight or CGAP portal, using the dcicutils project_utils mechanism.

from ..mime_types import MIME_TYPE_HTML, MIME_TYPE_JSON, MIME_TYPE_LD_JSON


class SnovaultProjectRenderers:
    def renderers_mime_types_supported(self):
        return [MIME_TYPE_JSON, MIME_TYPE_HTML, MIME_TYPE_LD_JSON]



import logging
from multiprocessing import cpu_count  # pylint: disable=no-name-in-module
from multiprocessing import Pool  # pylint: disable=no-name-in-module


log = logging.getLogger(__name__)


class ParallelTask(object):
    def __init__(self, task_func, num_cpu=None, no_parallel=False):
        """
        Args:
          - task_func (callable): Task to run on each work item. Must be a
              global function, or instance of a global class, due to
              multiprocessing's limitations.
          - no_parallel (bool): If true, run everything in the main thread.
        """
        self.task_func = task_func
        self.no_parallel = no_parallel
        self.num_cpu = num_cpu or cpu_count() - 1

    def run(self, items, chunk_size=1):
        """Run task in parallel on a list of work items.

        Uses multiprocessing in order to avoid Python's GIL.
        """
        if not self.no_parallel:
            with Pool(self.num_cpu) as pool:
                for res in pool.imap(self.task_func, items, chunk_size):
                    yield res
        else:
            for res in map(self.task_func, items):
                yield res



class SnovaultProjectSchemaViews:

    def get_submittable_item_names(self):
        """
        Optional project dependent list of item names in snake case format to include as submittable
        """
        return []
 
    def get_prop_for_submittable_items(self):
        """
        optional property name when if found as a top level property in a schema indicates it is submittable
        """
        return ""

    def get_properties_for_exclusion(self):
        """
        optional properties that should be omitted from submittable fields
        """
        return []

    def get_attributes_for_exclusion(self):
        """
        optional attributes of a property that if found exclude the property
        from submmittable propertie
        eg. {'permission':['restricted_fields'], 'calculatedProperty': [True]}
        """
        return {}



import json


class IngestionMessage:
    """
    Wrapper for raw ingestion message from SQS. Extracts ingestion type and uuid
    for easy/unified access. This is the type (of the first argument) expected by
    the ingestion_message_handler decorator. Note that the ingestion type name
    string is trimmed and treated as case-insensitive.
    """
    def __init__(self, raw_message: dict) -> None:
        self.body = json.loads(raw_message["Body"]) or {}
        self.uuid = self.body["uuid"] or ""
        self.type = self.body.get("ingestion_type", "vcf").strip().lower()

    def is_type(self, value: str) -> bool:
        return isinstance(value, str) and self.type == value.lower()



"""
Collection for objects related to ingestion submissions.
"""

import boto3
import contextlib
import io
import json
import logging
from typing import Optional
import traceback

from dcicutils.misc_utils import PRINT  # , ignored, check_true, VirtualApp
from snovault import collection, load_schema
# from pyramid.request import Request
# from pyramid.security import Allow, Deny, Everyone
from .base import (
    Item,
    # TODO: Maybe collect all these permission styles into a single file, give them symbolic names,
    #       and permit only the symbolic names to be used in each situation so we can curate a full inventory of modes.
    #       -kmp 26-Jul-2020
    # Ticket C4-332
    # ALLOW_PROJECT_MEMBER_ADD_ACL,
    # ONLY_ADMIN_VIEW_ACL,
)
from ..util import (
    debuglog, beanstalk_env_from_registry, create_empty_s3_file, s3_local_file, s3_output_stream,  # subrequest_item_creation,
    make_vapp_for_ingestion,  # vapp_for_email,
)
from ..ingestion.common import metadata_bundles_bucket  # , get_parameter


# ALLOW_SUBMITTER_VIEW_ACL = (
#     # TODO: There is an issue here where we want a logged in user remotely only to view this
#     #       but if we are proxying for them internall we want to be able to view OR edit.
#     #       There is never reason for a user outside the system to update this status. -kmp 26-Jul-2020
#     []  # Special additional permissions might go here.
#     + ALLOW_PROJECT_MEMBER_ADD_ACL  # Is this right? See note above.
#     + ONLY_ADMIN_VIEW_ACL    # Slightly misleading name. Allows admins to edit, too, actually. But only they can view.
# )


class SubmissionFolio:

    INGESTION_SUBMISSION_URI = '/IngestionSubmission'

    def __init__(self, *, vapp, ingestion_type, submission_id, log=None):
        self.vapp = vapp
        self._admin_vapp = make_vapp_for_ingestion(app=vapp.app)
        self.ingestion_type = ingestion_type
        self.log = log or logging
        self.bs_env = beanstalk_env_from_registry(vapp.app.registry)
        self.bucket = metadata_bundles_bucket(vapp.app.registry)
        self.s3_client = boto3.client('s3')
        self.other_details = {}
        self.outcome = 'unknown'
        self.submission_id = submission_id
        # These next two are initialized later by s3 lookup, and the result is cached here.
        # In particular, the values will be made available in time for the body of 'with folio.processing_context(...)'
        # Setting them to None here makes PyCharm and other code analysis tools happier in knowing
        # that accesses to these instance variables are legit. -kmp 27-Aug-2020
        self.object_name = None
        self.parameters = None
        self.resolution = None
        self.s3_encrypt_key_id = None  # This is overridden based on manifest later

    def __str__(self):
        return "<SubmissionFolio(%s) %s>" % (self.ingestion_type, self.submission_id)

    @classmethod
    def make_submission_uri(cls, submission_id):
        return "/ingestion-submissions/" + submission_id

    @property
    def submission_uri(self):
        return self.make_submission_uri(self.submission_id)

    def patch_item(self, **kwargs):
        res = self._admin_vapp.patch_json(self.submission_uri, kwargs)
        [item] = res.json['@graph']
        debuglog(json.dumps(item))
        return item

    def get_item(self):
        res = self._admin_vapp.get(self.submission_uri)
        [item] = res.json['@graph']
        return item

    def note_additional_datum(self, key, from_dict, from_key=None, default=None):
        self.other_details['additional_data'] = additional_data = (
            self.other_details.get('additional_data', {})
        )
        additional_data[key] = from_dict.get(from_key or key, default)

    @contextlib.contextmanager
    def s3_output(self, key_name, key_type='txt'):
        key = "%s/%s%s" % (self.submission_id, key_name, "" if key_name.endswith(f".{key_type}") else f".{key_type}")
        self.resolution[key_name] = key
        with s3_output_stream(self.s3_client, bucket=self.bucket, key=key,
                              s3_encrypt_key_id=self.s3_encrypt_key_id) as fp:
            yield fp

    @contextlib.contextmanager
    def s3_input(self, bucket: Optional[str] = None, key: Optional[str] = None):
        if not bucket:
            bucket = self.bucket
        if not key:
            key = self.object_name
        with s3_local_file(s3_client=self.s3_client, bucket=bucket, key=key) as filename:
            with io.open(filename, "r") as fp:
                yield fp

    def get_s3_input_json(self, bucket: Optional[str] = None, key: Optional[str] = None):
        with self.s3_input(bucket=bucket, key=key) as fp:
            return json.load(fp)

    def fail(self):
        self.outcome = 'failure'

    def succeed(self):
        self.outcome = 'success'

    def is_done(self):
        return self.outcome != 'unknown'

    @contextlib.contextmanager
    def processing_context(self):

        self.log.info("Processing {submission_id} as {ingestion_type}."
                      .format(submission_id=self.submission_id, ingestion_type=self.ingestion_type))

        submission_id = self.submission_id
        manifest_key = "%s/manifest.json" % submission_id
        response = self.s3_client.get_object(Bucket=self.bucket, Key=manifest_key)
        manifest = json.load(response['Body'])

        s3_encrypt_key_id = manifest.get("s3_encrypt_key_id")

        self.object_name = object_name = manifest['object_name']
        self.parameters = parameters = manifest['parameters']
        self.s3_encrypt_key_id = manifest['s3_encrypt_key_id']
        email = manifest['email']

        debuglog(submission_id, "object_name:", object_name)
        debuglog(submission_id, "parameters:", parameters)
        debuglog(submission_id, "s3_encrypt_key_id:", s3_encrypt_key_id)

        started_key = "%s/started.txt" % submission_id
        create_empty_s3_file(self.s3_client, bucket=self.bucket, key=started_key, s3_encrypt_key_id=s3_encrypt_key_id)

        # PyCharm thinks this is unused. -kmp 26-Jul-2020
        # data_stream = submission.s3_client.get_object(Bucket=submission.bucket,
        #                                               Key="%s/manifest.json" % submission_id)['Body']

        resolution = {
            "data_key": object_name,
            "manifest_key": manifest_key,
            "started_key": started_key,
        }

        try:

            other_keys = {}
            if email:
                other_keys['submitted_by'] = email

            self.patch_item(submission_id=submission_id,
                            object_name=object_name,
                            parameters=parameters,
                            processing_status={"state": "processing"},
                            **other_keys)

            self.resolution = resolution

            yield resolution

            if not self.is_done():
                self.succeed()

            self.patch_item(processing_status={"state": "done", "outcome": self.outcome, "progress": "complete"},
                            **self.other_details)

        except Exception as e:

            resolution["traceback_key"] = traceback_key = "%s/traceback.txt" % submission_id
            with s3_output_stream(self.s3_client, bucket=self.bucket, key=traceback_key,
                                  s3_encrypt_key_id=s3_encrypt_key_id) as fp:
                traceback.print_exc(file=fp)

            resolution["error_type"] = e.__class__.__name__
            resolution["error_message"] = str(e)

            self.patch_item(
                errors=["%s: %s" % (e.__class__.__name__, e)],
                processing_status={"state": "done", "outcome": "error", "progress": "incomplete"},
                **self.other_details
            )

        with s3_output_stream(self.s3_client,
                              bucket=self.bucket,
                              key="%s/resolution.json" % submission_id,
                              s3_encrypt_key_id=s3_encrypt_key_id) as fp:
            PRINT(json.dumps(resolution, indent=2), file=fp)

    def process_standard_bundle_results(self, bundle_result: dict, s3_only: bool = False) -> None:
        """
        If the given bundle_result contains either a result, post_output, or upload_info property,
        then writes the contents of that property to an S3 key with either the name submission.json,
        post_output.txt, or upload_info.txt, respectively; and in the bucket name self.bucket.

        Additionally, for these three properties (result, post_output, upload_info), add them to the
        self.other_details["additional_data"] property of this SubmissionFolio object, so that this
        data will ultimately be written to the database (for the IngestionSubmission object).

        HOWEVER, if the s3_only (False by default) argument is True then the given result will
        NOT be added to the other_details["additional_data"] of this SubmissionFolio object,
        rather the given result will ONLY be written to S3. This is to prevent potentially
        larger amounts of data from ultimately being written to the database (for the
        IngestionSubmission object), but rather have them stored ONLY in S3.
        """

        # Next several files are created only if relevant.

        if bundle_result.get('result'):
            with self.s3_output(key_name='submission.json', key_type='json') as fp:
                PRINT(json.dumps(bundle_result['result'], indent=2), file=fp)
                if not s3_only:
                    self.note_additional_datum('result', from_dict=bundle_result, default={})

        if bundle_result.get('validation_output'):
            with self.s3_output(key_name='summary.json', key_type='json') as fp:
                PRINT(json.dumps(bundle_result['validation_output'], indent=2), file=fp)
                if not s3_only:
                    self.note_additional_datum('validation_output', from_dict=bundle_result, default={})

        if bundle_result.get('post_output'):
            with self.s3_output(key_name='submission_response') as fp:
                self.show_report_lines(bundle_result['post_output'], fp)
                if not s3_only:
                    self.note_additional_datum('post_output', from_dict=bundle_result, default=[])

        if bundle_result.get('upload_info'):
            with self.s3_output(key_name='upload_info') as fp:
                PRINT(json.dumps(bundle_result['upload_info'], indent=2), file=fp)
                if not s3_only:
                    self.note_additional_datum('upload_info', from_dict=bundle_result, default=[])

    @staticmethod
    def show_report_lines(lines, fp, default="Nothing to report."):
        for line in lines or ([default] if default else []):
            try:
                PRINT(line, file=fp)
            except UnicodeEncodeError:
                ascii_line = line.encode(
                    encoding="ascii", errors="backslashreplace"
                ).decode(encoding="ascii")
                PRINT(ascii_line, file=fp)


@collection(
    name='ingestion-submissions',
    # acl=ALLOW_SUBMITTER_VIEW_ACL,
    unique_key='object_name',
    properties={
        'title': 'Ingestion Submissions',
        'description': 'List of Ingestion Submissions',
    })
class IngestionSubmission(Item):
    """The IngestionSubmission class that holds info on requests to ingest data."""

    item_type = 'ingestion_submission'
    schema = load_schema('snovault:schemas/ingestion_submission.json')
    # embedded_list = [...] + Item.embedded_list

    schema_json = json.loads(json.dumps(schema))

    @classmethod
    def supports_type(cls, ingestion_type: str) -> bool:
        return ingestion_type in cls.schema_json["properties"]["ingestion_type"]["enum"]
